package de.ozgcloud.xta.client.core;

import jakarta.xml.ws.Holder;

import de.ozgcloud.xta.client.mapper.RequestMapper;
import de.ozgcloud.xta.client.mapper.ResponseMapper;
import de.ozgcloud.xta.client.model.XtaIdentifier;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaDataListing;
import de.ozgcloud.xta.client.model.XtaTransportReport;
import genv3.de.xoev.transport.xta.x211.InvalidMessageIDException;
import genv3.de.xoev.transport.xta.x211.MessageSchemaViolationException;
import genv3.de.xoev.transport.xta.x211.MessageVirusDetectionException;
import genv3.de.xoev.transport.xta.x211.ParameterIsNotValidException;
import genv3.de.xoev.transport.xta.x211.PermissionDeniedException;
import genv3.de.xoev.transport.xta.x211.SyncAsyncException;
import genv3.de.xoev.transport.xta.x211.XTAWSTechnicalProblemException;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType;
import genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData;
import lombok.Builder;
import lombok.Getter;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
@Builder
@Getter
public class WrappedXtaService {

	private final XtaPortTriple ports;
	private final RequestMapper requestMapper;
	private final ResponseMapper responseMapper;

	public void checkAccountActive(XtaIdentifier clientIdentifier) throws XTAWSTechnicalProblemException, PermissionDeniedException {
		ports.managementPort().checkAccountActive(requestMapper.mapPartyTypeFromIdentifier(clientIdentifier));
	}

	public XtaMessage getMessage(String messageId, XtaIdentifier clientIdentifier)
			throws XTAWSTechnicalProblemException, PermissionDeniedException, InvalidMessageIDException {
		final Holder<MessageMetaData> messageMetaDataHolder = new Holder<>();
		var genericContentContainer = ports.msgBoxPort().getMessage(
				requestMapper.mapMsgBoxFetchRequest(messageId),
				requestMapper.mapPartyTypeFromIdentifier(clientIdentifier),
				messageMetaDataHolder,
				null
		);
		return responseMapper.mapXtaMessage(
				genericContentContainer,
				messageMetaDataHolder.value
		);
	}

	public XtaMessageMetaDataListing getStatusList(XtaIdentifier clientIdentifier, int maxListItems)
			throws XTAWSTechnicalProblemException, PermissionDeniedException {
		final Holder<MsgBoxResponseType> fetchResponseHeader = new Holder<>();
		var msgStatusListType = ports.msgBoxPort().getStatusList(
				requestMapper.mapMsgBoxStatusListRequestTypeFromMaxListItems(maxListItems),
				requestMapper.mapPartyTypeFromIdentifier(clientIdentifier),
				fetchResponseHeader
		);
		return responseMapper.mapXtaMessageMetaListing(msgStatusListType, fetchResponseHeader.value);
	}

	public void sendMessage(XtaMessage message)
			throws SyncAsyncException, XTAWSTechnicalProblemException, MessageVirusDetectionException, MessageSchemaViolationException,
			PermissionDeniedException, ParameterIsNotValidException {
		ports.sendPort().sendMessage(
				requestMapper.mapGenericContentContainer(message),
				requestMapper.mapMessageMetaDataFromXtaMessageMetaData(message.metaData()),
				null
		);
	}

	public String createMessageId(XtaIdentifier clientIdentifier) throws XTAWSTechnicalProblemException, PermissionDeniedException {
		return ports.managementPort().createMessageId(
				requestMapper.mapPartyTypeFromIdentifier(clientIdentifier)).getValue();
	}

	public Boolean lookupService(String service, XtaIdentifier reader, XtaIdentifier clientIdentifier)
			throws XTAWSTechnicalProblemException, PermissionDeniedException, ParameterIsNotValidException {
		return responseMapper.isServiceAvailable(
				ports.managementPort().lookupService(
						requestMapper.mapLookupServiceRequest(service, reader),
						requestMapper.mapPartyTypeFromIdentifier(clientIdentifier))
		);
	}

	public void close(String messageId, XtaIdentifier clientIdentifier)
			throws XTAWSTechnicalProblemException, PermissionDeniedException, InvalidMessageIDException {
		ports.msgBoxPort().close(
				requestMapper.mapMsgBoxCloseRequestType(messageId),
				requestMapper.mapPartyTypeFromIdentifier(clientIdentifier)
		);
	}

	public XtaTransportReport getTransportReport(String messageId, XtaIdentifier clientIdentifier)
			throws XTAWSTechnicalProblemException, PermissionDeniedException, InvalidMessageIDException {
		return responseMapper.mapXtaTransportReport(
				ports.managementPort().getTransportReport(
						requestMapper.mapAttributedURIType(messageId),
						requestMapper.mapPartyTypeFromIdentifier(clientIdentifier)
				));
	}

}
