package de.ozgcloud.xta.client.core;

import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;
import de.ozgcloud.xta.client.exception.XtaClientException;
import genv3.de.xoev.transport.xta.x211.ExceptionType;
import lombok.Builder;
import lombok.RequiredArgsConstructor;

@Builder
@RequiredArgsConstructor
public class XtaExceptionHandler {

	static final String UNEXPECTED_ERROR_MESSAGE = "An unexpected error occurred. Please report this to the xta-client maintainers.";

	public XtaClientException deriveXtaClientException(RuntimeException exception) {
		if (exception instanceof XtaClientRuntimeException xtaClientRuntimeException) {
			return deriveXtaClientExceptionFromClientRuntimeException(xtaClientRuntimeException);
		}
		return new XtaClientException(UNEXPECTED_ERROR_MESSAGE, exception);
	}

	XtaClientException deriveXtaClientExceptionFromClientRuntimeException(XtaClientRuntimeException exception) {
		var cause = exception.getCause();
		var detailMessageLines = getDetailLines(cause);
		var message = Stream.concat(
				Stream.of(exception.getMessage()),
				detailMessageLines
		).collect(Collectors.joining("\n"));
		return new XtaClientException(message, cause);
	}

	Stream<String> getDetailLines(Throwable cause) {
		return Optional.ofNullable(cause)
				.filter(Exception.class::isInstance)
				.map(Exception.class::cast)
				.map(this::deriveDetailLinesFromException)
				.orElse(Stream.empty());
	}

	Stream<String> deriveDetailLinesFromException(Exception exception) {
		return getExceptionType(exception)
				.map(ExceptionType::getErrorCode)
				.map(num -> Stream.of(exception.getMessage(), "[%s] %s".formatted(num.getCode(), num.getName())))
				.orElse(Stream.empty());
	}

	private Optional<ExceptionType> getExceptionType(Exception exception) {
		return getPrivateFieldValue(exception, "faultInfo")
				.filter(ExceptionType.class::isInstance)
				.map(ExceptionType.class::cast);
	}

	private Optional<Object> getPrivateFieldValue(Object object, String fieldName) {
		try {
			var field = object.getClass().getDeclaredField(fieldName);
			field.setAccessible(true);
			return Optional.of(field.get(object));
		} catch (NoSuchFieldException | IllegalAccessException | RuntimeException e) {
			return Optional.empty();
		}
	}

}
