package de.ozgcloud.xta.client.core;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;

import javax.net.ssl.SSLContext;

import org.apache.cxf.configuration.jsse.TLSClientParameters;
import org.apache.hc.core5.ssl.SSLContextBuilder;

import de.ozgcloud.xta.client.config.XtaClientConfig;
import de.ozgcloud.xta.client.exception.XtaClientInitializationException;
import lombok.Builder;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;

@Builder
@RequiredArgsConstructor
@Slf4j
public class XtaTLSClientParametersFactory {

	static final String TLS_PROTOCOL = "TLSv1.2";

	private final XtaClientConfig config;

	public TLSClientParameters create() throws XtaClientInitializationException {
		log.debug("[createXtaTLsParameters] Using config: {}", config);
		var sslContext = createXtaSslContext();

		var parameters = new TLSClientParameters();
		parameters.setSSLSocketFactory(sslContext.getSocketFactory());
		return parameters;
	}

	public SSLContext createXtaSslContext() throws XtaClientInitializationException {
		try {
			var sslContextBuilder = createSSLContextBuilder();
			var clientCertKeystore = config.getClientCertKeystore();
			if (clientCertKeystore != null) {
				sslContextBuilder.loadKeyMaterial(loadStore(clientCertKeystore), clientCertKeystore.getPassword());
			}
			var trustStore = config.getTrustStore();
			if (trustStore != null) {
				sslContextBuilder.loadTrustMaterial(loadStore(trustStore), null);
			}
			return sslContextBuilder.build();
		} catch (NoSuchAlgorithmException | KeyStoreException | KeyManagementException | UnrecoverableKeyException | IOException |
				CertificateException e) {
			throw new XtaClientInitializationException("Failed to create SSL context: " + e.getMessage(), e);
		}
	}

	SSLContextBuilder createSSLContextBuilder() {
		return SSLContextBuilder.create()
				.setProtocol(TLS_PROTOCOL);
	}

	KeyStore loadStore(final XtaClientConfig.KeyStore keystore)
			throws IOException, KeyStoreException, CertificateException, NoSuchAlgorithmException {
		var keyStore = KeyStore.getInstance(keystore.getType());
		try (var is = new ByteArrayInputStream(keystore.getContent())) {
			keyStore.load(is, keystore.getPassword());
		}
		return keyStore;
	}
}
