package de.ozgcloud.xta.client.xdomea;

import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.xpath.XPathFactory;
import javax.xml.xpath.XPathFactoryConfigurationException;

import org.apache.cxf.common.xmlschema.LSInputImpl;
import org.mapstruct.factory.Mappers;
import org.w3c.dom.ls.LSResourceResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;
import de.ozgcloud.xta.client.xdomea.mapper.MetadataMapper;
import de.ozgcloud.xta.client.xdomea.reader.XdomeaValueReaderFactory;
import de.ozgcloud.xta.client.xdomea.reader.ZipFileEntryReader;
import de.ozgcloud.xta.client.xdomea.reader.ZipFileEntryReaderFactory;
import lombok.Builder;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;

@Builder
@RequiredArgsConstructor
@Slf4j
public class XdomeaXtaMessageCreatorFactory {

	static final String DISALLOW_DOCTYPE_DECL_FEATURE_URL = "http://apache.org/xml/features/disallow-doctype-decl";
	static final List<String> SCHEMA_RESOURCES = List.of(
			"xdomea/240/xdomea-Nachrichten-VBS.xsd",
			"xdomea/240/xdomea-Nachrichten-DMS.xsd",
			"xdomea/300/xdomea.xsd"
	);

	static final Map<String, String> CACHED_SCHEMA_RESOURCES = Map.of(
			"http://xoev.de/schemata/code/1_0/xoev-code.xsd", "xdomea/xoev-code.xsd",
			"http://xoev.de/latinchars/1_1/datatypes/latinchars.xsd", "xdomea/latinchars.xsd"
	);

	private final XdomeaValueReaderFactory xdomeaValueReaderFactory;
	private final ZipFileEntryReader zipFileEntryReader;
	private final DocumentBuilderFactory documentBuilderFactory;
	private final MetadataMapper metadataMapper;
	private final XdomeaMetaDataValidatorFactory xdomeaMetaDataValidatorFactory;
	private final SchemaFactory schemaFactory;
	private final ClassLoader classLoader;
	private final ErrorHandler errorHandler;

	public static XdomeaXtaMessageCreatorFactory createInstance() {
		var zipEntryReaderFactory = ZipFileEntryReaderFactory.builder().build();
		var zipEntryReader = zipEntryReaderFactory.create();
		return XdomeaXtaMessageCreatorFactory.builder()
				.xdomeaValueReaderFactory(XdomeaValueReaderFactory.builder()
						.xPathFactory(createXPathFactory())
						.build())
				.zipFileEntryReader(zipEntryReader)
				.documentBuilderFactory(DocumentBuilderFactory.newInstance())
				.metadataMapper(Mappers.getMapper(MetadataMapper.class))
				.xdomeaMetaDataValidatorFactory(XdomeaMetaDataValidatorFactory.builder()
						.zipFileEntryReader(zipEntryReader)
						.build())
				.schemaFactory(SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI))
				.classLoader(XdomeaXtaMessageCreatorFactory.class.getClassLoader())
				.errorHandler(new CustomErrorHandler())
				.build();
	}

	private static XPathFactory createXPathFactory() {
		var xPathFactory = XPathFactory.newInstance();
		try {
			xPathFactory.setFeature(XMLConstants.FEATURE_SECURE_PROCESSING, Boolean.TRUE);
			return xPathFactory;
		} catch (XPathFactoryConfigurationException e) {
			throw new XtaClientRuntimeException("Failed to configure xpath factory!", e);
		}
	}

	static class CustomErrorHandler implements ErrorHandler {
		@Override
		public void warning(SAXParseException e) {
			log.warn("Schema warning!", e);
		}

		@Override
		public void error(SAXParseException e) throws SAXException {
			throw e;
		}

		@Override
		public void fatalError(SAXParseException e) throws SAXException {
			throw e;
		}
	}

	LSResourceResolver createResourceResolver() {
		return (type, namespaceURI, publicId, systemId, baseURI) -> Optional.ofNullable(CACHED_SCHEMA_RESOURCES.get(systemId))
				.map(resourceName -> new LSInputImpl(publicId, systemId, classLoader.getResourceAsStream(resourceName)))
				.orElse(null);
	}

	StreamSource createSchemaStreamSource(String resourcePath) {
		var resource = Objects.requireNonNull(classLoader.getResource(resourcePath), "Resource not found: " + resourcePath);
		return new StreamSource(resource.toExternalForm());
	}

	StreamSource[] createSchemaStreamSources() {
		return SCHEMA_RESOURCES.stream()
				.peek(resourcePath -> log.debug("Initializing schema '{}' ...", resourcePath))
				.map(this::createSchemaStreamSource)
				.toArray(StreamSource[]::new);
	}

	Schema createSchema() {
		try {
			schemaFactory.setResourceResolver(createResourceResolver());
			return schemaFactory.newSchema(createSchemaStreamSources());
		} catch (SAXException e) {
			throw new XtaClientRuntimeException("Failed to initialize schema", e);
		}
	}

	private DocumentBuilder createDocumentBuilder() {
		try {
			documentBuilderFactory.setFeature(DISALLOW_DOCTYPE_DECL_FEATURE_URL, true);
			documentBuilderFactory.setNamespaceAware(true);
			documentBuilderFactory.setSchema(createSchema());
			log.debug("Creating new document builder...");
			var documentBuilder = documentBuilderFactory.newDocumentBuilder();
			documentBuilder.setErrorHandler(errorHandler);
			return documentBuilder;
		} catch (ParserConfigurationException e) {
			throw new XtaClientRuntimeException("Failed to configure document builder!", e);
		}
	}

	public XdomeaXtaMessageCreator create() {
		return XdomeaXtaMessageCreator.builder()
				.documentBuilder(createDocumentBuilder())
				.xdomeaValueReader(xdomeaValueReaderFactory.create())
				.zipFileEntryReader(zipFileEntryReader)
				.metadataMapper(metadataMapper)
				.metaDataValidator(xdomeaMetaDataValidatorFactory.create())
				.build();
	}
}
