package de.ozgcloud.xta.client.xdomea.reader;

import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.w3c.dom.Document;

import de.ozgcloud.xta.client.exception.XtaClientException;
import lombok.Builder;
import lombok.RequiredArgsConstructor;

@Builder
@RequiredArgsConstructor
public class XdomeaValueReader {

	static final String PROCESS_ID_XPATH = getXpath("Kopf", "ProzessID");
	static final String AUTHOR_ID_SUFFIX_XPATH = getXpath("Kopf", "Absender", "Behoerdenschluessel|Kennung", "code");
	static final String AUTHOR_ID_PREFIX_XPATH = getXpath("Kopf", "Absender", "Praefix", "code");
	static final String READER_ID_SUFFIX_XPATH = getXpath("Kopf", "Empfaenger", "Behoerdenschluessel|Kennung", "code");
	static final String READER_ID_PREFIX_XPATH = getXpath("Kopf", "Empfaenger", "Praefix", "code");
	static final String MESSAGE_TYPE_ID_SUFFIX_XPATH = getXpath("Kopf", "Nachrichtentyp", "code");
	static final String PRIMARY_DOCUMENT_NAME_XPATH = getXpath("Primaerdokument", "Dateiname");

	static String getXpath(String... hierarchy) {
		return Arrays.stream(hierarchy)
				.map(localNames -> Arrays.stream(localNames.split("\\|"))
						.map(localName -> "local-name() = '" + localName + "'")
						.collect(Collectors.joining(" or ")))
				.map(predicate -> "//*[" + predicate + "]")
				.collect(Collectors.joining());
	}

	private final Map<String, XmlValueReader> xmlValueReaders;

	public XdomeaXmlValues readValues(Document xdomeaXmlDocument) throws XtaClientException {
		return XdomeaXmlValues.builder()
				.processId(readRequiredValue(xdomeaXmlDocument, PROCESS_ID_XPATH))
				.messageTypeCode(readRequiredValue(xdomeaXmlDocument, MESSAGE_TYPE_ID_SUFFIX_XPATH))
				.authorIdSuffix(readRequiredValue(xdomeaXmlDocument, AUTHOR_ID_SUFFIX_XPATH))
				.authorIdPrefix(readRequiredValue(xdomeaXmlDocument, AUTHOR_ID_PREFIX_XPATH))
				.readerIdSuffix(readRequiredValue(xdomeaXmlDocument, READER_ID_SUFFIX_XPATH))
				.readerIdPrefix(readRequiredValue(xdomeaXmlDocument, READER_ID_PREFIX_XPATH))
				.primaryDocumentNames(readPrimaryDocumentNames(xdomeaXmlDocument))
				.build();
	}

	List<String> readPrimaryDocumentNames(Document xdomeaXmlDocument) {
		return getXmlValueReader(PRIMARY_DOCUMENT_NAME_XPATH)
				.readNonEmptyTexts(xdomeaXmlDocument)
				.toList();
	}

	String readRequiredValue(Document xdomeaXmlDocument, String xpathString) throws XtaClientException {
		return getXmlValueReader(xpathString)
				.readNonEmptyTexts(xdomeaXmlDocument)
				.findFirst()
				.orElseThrow(() -> new XtaClientException("Required value " + xpathString + " not found in xdomea xml document!"));
	}

	XmlValueReader getXmlValueReader(String xpath) {
		return xmlValueReaders.get(xpath);
	}

}
