package de.ozgcloud.xta.client;

import static de.ozgcloud.xta.client.factory.MessageMetaDataTestFactory.*;
import static de.ozgcloud.xta.client.factory.XtaClientConfigTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.math.BigInteger;
import java.util.List;
import java.util.Optional;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.xta.client.config.XtaClientConfig;
import de.ozgcloud.xta.client.core.XtaClientService;
import de.ozgcloud.xta.client.core.XtaExceptionHandler;
import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;
import de.ozgcloud.xta.client.factory.ClientRuntimeExceptionTestFactory;
import de.ozgcloud.xta.client.factory.XtaClientExceptionTestFactory;
import de.ozgcloud.xta.client.factory.XtaMessageMetaDataListingTestFactory;
import de.ozgcloud.xta.client.factory.XtaMessageMetaDataTestFactory;
import de.ozgcloud.xta.client.factory.XtaMessageTestFactory;
import de.ozgcloud.xta.client.factory.XtaTransportReportTestFactory;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;
import de.ozgcloud.xta.client.model.XtaMessageMetaDataListing;
import de.ozgcloud.xta.client.model.XtaMessageStatus;
import de.ozgcloud.xta.client.model.XtaTransportReport;
import lombok.SneakyThrows;

class XtaClientTest {

	@Mock
	private XtaClientService service;

	@Mock
	private XtaClientConfig config;

	@Mock
	private XtaExceptionHandler exceptionHandler;

	@Spy
	@InjectMocks
	private XtaClient client;

	private XtaMessageMetaDataListing listing;

	private List<XtaTransportReport> transportReports;

	private XtaMessage message;

	@BeforeEach
	void setup() {
		message = XtaMessageTestFactory.create();
		listing = XtaMessageMetaDataListingTestFactory.create();
		transportReports = listing.messages().stream()
				.map(messageMetaData -> XtaTransportReportTestFactory.createBuilder()
						.metaData(messageMetaData)
						.build())
				.toList();
	}

	@DisplayName("fetch messages")
	@Nested
	class TestFetchMessages {

		@Mock
		private Consumer<XtaMessage> processMessage;

		@Mock
		private XtaTransportReport transportReport;

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			doReturn(List.of(transportReport)).when(client).fetchMessagesRaw(processMessage);

			var result = client.fetchMessages(processMessage);

			assertThat(result).containsExactly(transportReport);
		}

		@DisplayName("should throw checked exception on runtime exception")
		@Test
		void shouldThrowCheckedExceptionOnRuntimeException() {
			var clientException = XtaClientExceptionTestFactory.create();
			var runTimeException = ClientRuntimeExceptionTestFactory.create();
			doThrow(runTimeException).when(client).fetchMessagesRaw(processMessage);
			when(exceptionHandler.deriveXtaClientException(runTimeException)).thenReturn(clientException);

			assertThatThrownBy(() -> client.fetchMessages(processMessage))
					.isEqualTo(clientException);
		}

	}

	@DisplayName("fetch messages raw")
	@Nested
	class TestFetchMessagesRaw {

		@Mock
		private Consumer<XtaMessage> processMessage;

		@Mock
		private FetchMessageParameter parameter;

		@Mock
		private XtaTransportReport transportReport;

		@BeforeEach
		void mock() {
			doReturn(List.of(SELF_IDENTIFIER)).when(config).getClientIdentifiers();
		}

		@DisplayName("with active account")
		@Nested
		class TestWithActiveAccount {

			@DisplayName("should fetch messages for client identifier")
			@Test
			void shouldFetchMessagesForClientIdentifier() {
				doReturn(true).when(service).checkAccountActive(SELF_IDENTIFIER);
				doReturn(parameter).when(client).initializeFetchMessageParameter(SELF_IDENTIFIER, processMessage);
				doReturn(Stream.of(transportReport)).when(client).fetchMessagesForClientIdentifier(parameter);

				var result = fetchMessages();

				assertThat(result).containsExactly(transportReport);
			}
		}

		@DisplayName("with inactive account")
		@Nested
		class TestWithInactiveAccount {

			@DisplayName("should fetch no messages")
			@Test
			void shouldFetchNoMessages() {
				doReturn(false).when(service).checkAccountActive(SELF_IDENTIFIER);

				var result = fetchMessages();

				assertThat(result).isEmpty();
			}
		}

		@SneakyThrows
		private List<XtaTransportReport> fetchMessages() {
			return client.fetchMessages(processMessage);
		}
	}

	@DisplayName("initialize fetch message parameter")
	@Nested
	class TestInitializeFetchMessageParameter {

		@Mock
		private Consumer<XtaMessage> processMessage;

		@DisplayName("should init empty viewed message ids")
		@Test
		void shouldInitEmptyViewedMessageIds() {
			var result = initializeFetchMessageParameter();

			assertThat(result.viewedMessageIds()).isEmpty();
		}

		@DisplayName("should set client identifier")
		@Test
		void shouldSetClientIdentifier() {
			var result = initializeFetchMessageParameter();

			assertThat(result.clientIdentifier()).isEqualTo(SELF_IDENTIFIER);
		}

		@DisplayName("should set process message")
		@Test
		void shouldSetProcessMessage() {
			var result = initializeFetchMessageParameter();

			assertThat(result.processMessage()).isEqualTo(processMessage);
		}

		FetchMessageParameter initializeFetchMessageParameter() {
			return client.initializeFetchMessageParameter(SELF_IDENTIFIER, processMessage);
		}
	}

	@DisplayName("fetch messages for client identifier")
	@Nested
	class TestFetchMessagesForClientIdentifier {

		@Mock
		private FetchMessageParameter parameter;

		@BeforeEach
		void mock() {
			when(parameter.clientIdentifier()).thenReturn(SELF_IDENTIFIER);
		}

		@DisplayName("with successful listing")
		@Nested
		class TestWithSuccessfulListing {

			@Mock
			private FetchMessageParameter nextParameter;

			@BeforeEach
			void mock() {
				when(service.getStatusList(SELF_IDENTIFIER)).thenReturn(Optional.of(listing));
				doReturn(transportReports).when(client).fetchMessagesForListing(listing.messages(), parameter);
			}

			@DisplayName("should return transport reports if no more messages available\"")
			@Test
			void shouldReturnTransportReportsIfNoMoreMessagesAvailable() {
				doReturn(false).when(client).checkMoreMessagesAvailable(listing, transportReports);

				var result = client.fetchMessagesForClientIdentifier(parameter).toList();

				assertThat(result)
						.extracting(XtaTransportReport::metaData)
						.extracting(XtaMessageMetaData::messageId)
						.containsExactly(MESSAGE_ID, MESSAGE_ID2, MESSAGE_ID3);
			}

			@DisplayName("should return concatenated transport reports if more messages available")
			@Test
			void shouldReturnConcatenatedTransportReportsIfMoreMessagesAvailable() {
				var messageId4 = "messageId4";
				var additionalTransportReport = XtaTransportReportTestFactory.createBuilder()
						.metaData(XtaMessageMetaDataTestFactory.createBuilder()
								.messageId(messageId4)
								.build())
						.build();
				doReturn(true).when(client).checkMoreMessagesAvailable(listing, transportReports);
				doReturn(nextParameter).when(parameter).withViewedMessageIdsFrom(listing.messages());
				doReturn(Stream.of(additionalTransportReport)).when(client).fetchMessagesForClientIdentifier(nextParameter);

				var result = client.fetchMessagesForClientIdentifier(parameter).toList();

				assertThat(result)
						.extracting(XtaTransportReport::metaData)
						.extracting(XtaMessageMetaData::messageId)
						.containsExactly(MESSAGE_ID, MESSAGE_ID2, MESSAGE_ID3, messageId4);
			}
		}

		@DisplayName("with missing listing")
		@Nested
		class TestWithMissingListing {

			@BeforeEach
			void mock() {
				when(service.getStatusList(SELF_IDENTIFIER)).thenReturn(Optional.empty());
			}

			@DisplayName("should return empty stream")
			@Test
			void shouldReturnEmptyStream() {
				var result = client.fetchMessagesForClientIdentifier(parameter).toList();

				assertThat(result).isEmpty();
			}
		}
	}

	@DisplayName("check more messages available")
	@Nested
	class TestCheckMoreMessagesAvailable {

		@DisplayName("with no extra pending messages")
		@Nested
		class TestWithNoExtraPendingMessages {
			@DisplayName("should return false")
			@Test
			void shouldReturnFalse() {
				doReturn(false).when(client).checkExtraPendingMessagesAvailable(listing);

				var result = client.checkMoreMessagesAvailable(listing, transportReports);

				assertThat(result).isFalse();
			}
		}

		@DisplayName("with extra pending messages but no message closed")
		@Nested
		class TestWithExtraPendingMessagesButNoMessageClosed {
			@DisplayName("should return false")
			@Test
			void shouldReturnFalse() {
				doReturn(true).when(client).checkExtraPendingMessagesAvailable(listing);
				doReturn(false).when(client).checkSomeMessageHasBeenClosed(transportReports);

				var result = client.checkMoreMessagesAvailable(listing, transportReports);

				assertThat(result).isFalse();
			}
		}
	}

	@DisplayName("check extra pending messages available")
	@Nested
	class TestCheckExtraPendingMessagesAvailable {
		@DisplayName("should return true if more pending than received")
		@Test
		void shouldReturnTrueIfMorePendingThanReceived() {
			var result = client.checkExtraPendingMessagesAvailable(listing);

			assertThat(result).isTrue();
		}

		@DisplayName("should return false if no more pending than received")
		@Test
		void shouldReturnFalseIfNoMorePendingThanReceived() {
			var listingWithNoExtraPendingMessages = XtaMessageMetaDataListingTestFactory.createBuilder()
					.pendingMessageCount(BigInteger.valueOf(listing.messages().size()))
					.build();

			var result = client.checkExtraPendingMessagesAvailable(listingWithNoExtraPendingMessages);

			assertThat(result).isFalse();
		}
	}

	@DisplayName("check some message has been closed")
	@Nested
	class TestCheckSomeMessageHasBeenClosed {
		@DisplayName("should return false if no message has green status")
		@Test
		void shouldReturnFalseIfNoMessageHasGreenStatus() {
			var result = client.checkSomeMessageHasBeenClosed(transportReports);

			assertThat(result).isFalse();
		}

		@DisplayName("should log warning if no message has been closed")
		@Test
		void shouldLogWarningIfNoMessageHasBeenClosed() {
			client.checkSomeMessageHasBeenClosed(transportReports);

			verify(client).logWarnForNoMessageClosed();
		}

		@DisplayName("should return true if some message has green status")
		@Test
		void shouldReturnTrueIfSomeMessageHasGreenStatus() {
			var transportReportsWithGreenStatus = Stream.concat(
					transportReports.stream(),
					Stream.of(XtaTransportReportTestFactory.createBuilder()
							.status(XtaMessageStatus.GREEN)
							.build())
			).toList();

			var result = client.checkSomeMessageHasBeenClosed(transportReportsWithGreenStatus);

			assertThat(result).isTrue();
		}
	}

	@DisplayName("fetch messages for listing")
	@Nested
	class TestFetchMessagesForListing {

		@Mock
		private FetchMessageParameter parameter;

		@Captor
		private ArgumentCaptor<XtaMessageMetaData> messageMetaDataArgumentCaptor;

		@DisplayName("with messages which should not be fetched")
		@Nested
		class TestWithMessagesWhichShouldNotBeFetched {

			@BeforeEach
			void mock() {
				doReturn(false).when(client).checkMessageShouldBeFetched(any(), any());
			}

			@DisplayName("should check if message has already been viewed")
			@Test
			void shouldCheckIfMessageHasAlreadyBeenViewed() {
				client.fetchMessagesForListing(listing.messages(), parameter);

				verify(client, times(3)).checkMessageShouldBeFetched(messageMetaDataArgumentCaptor.capture(), eq(parameter));
				assertThat(messageMetaDataArgumentCaptor.getAllValues())
						.containsExactlyElementsOf(listing.messages());
			}
		}

		@DisplayName("with messages which should be fetched")
		@Nested
		class TestWithMessagesWhichShouldBeFetched {

			@BeforeEach
			void mock() {
				doReturn(true).when(client).checkMessageShouldBeFetched(any(), any());
				doAnswer(args -> {
					XtaMessageMetaData metaData = args.getArgument(0);
					assertThat(metaData).isNotNull();
					return getTestTransportReportByMessageId(metaData.messageId());
				}).when(client).fetchMessage(any(), any());
			}

			private Optional<XtaTransportReport> getTestTransportReportByMessageId(String messageId) {
				assertThat(messageId).isNotNull();
				return transportReports.stream().filter(d -> {
					assertThat(d.metaData().messageId()).isNotNull();
					return d.metaData().messageId().equals(messageId);
				}).findFirst();
			}

			@DisplayName("should return transport reports")
			@Test
			void shouldReturnTransportReports() {
				var result = client.fetchMessagesForListing(listing.messages(), parameter);

				assertThat(result).containsExactlyElementsOf(transportReports);
			}
		}
	}

	@DisplayName("check message should be fetched")
	@Nested
	class TestCheckMessageShouldBeFetched {
		@Mock
		private FetchMessageParameter parameter;

		@Mock
		private XtaMessageMetaData messageMetaData;

		@DisplayName("should return false if message already viewed")
		@Test
		void shouldReturnFalseIfMessageAlreadyViewed() {
			when(parameter.hasMessageAlreadyBeenViewed(messageMetaData)).thenReturn(true);

			var result = checkMessageShouldBeFetched();

			assertThat(result).isFalse();
		}

		@DisplayName("with message not viewed")
		@Nested
		class TestWithMessageNotViewed {

			@BeforeEach
			void mock() {
				when(parameter.hasMessageAlreadyBeenViewed(messageMetaData)).thenReturn(false);
			}

			@DisplayName("should return false if message is not supported")
			@Test
			void shouldReturnFalseIfMessageIsNotSupported() {
				doReturn(false).when(client).isMessageSupported(messageMetaData);

				var result = checkMessageShouldBeFetched();

				assertThat(result).isFalse();
			}

			@DisplayName("should return true if message is supported")
			@Test
			void shouldReturnTrueIfMessageIsSupported() {
				doReturn(true).when(client).isMessageSupported(messageMetaData);

				var result = checkMessageShouldBeFetched();

				assertThat(result).isTrue();
			}
		}

		private boolean checkMessageShouldBeFetched() {
			return client.checkMessageShouldBeFetched(messageMetaData, parameter);
		}
	}

	@DisplayName("is message supported")
	@Nested
	class TestIsMessageSupported {

		@Mock
		private Predicate<XtaMessageMetaData> isSupportedPredicate;

		@Mock
		private XtaMessageMetaData messageMetaData;

		@DisplayName("should return true if predicate is null")
		@Test
		void shouldReturnTrueIfPredicateIsNull() {
			when(config.getIsMessageSupported()).thenReturn(null);

			var result = client.isMessageSupported(messageMetaData);

			assertThat(result).isTrue();
		}

		@DisplayName("should return predicate value")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldReturnPredicateValue(boolean value) {
			when(config.getIsMessageSupported()).thenReturn(isSupportedPredicate);
			when(isSupportedPredicate.test(messageMetaData)).thenReturn(value);

			var result = client.isMessageSupported(messageMetaData);

			assertThat(result).isEqualTo(value);
		}
	}

	@DisplayName("fetch message")
	@Nested
	class TestFetchMessage {

		@Mock
		private XtaMessageMetaData messageMetaData;

		@Mock
		private FetchMessageParameter parameter;

		@Mock
		private XtaMessage message;

		@Mock
		private XtaTransportReport transportReport;

		@DisplayName("should return transport report")
		@Test
		void shouldReturnTransportReport() {
			when(service.getMessage(messageMetaData)).thenReturn(Optional.of(message));
			doReturn(Optional.of(transportReport)).when(client).processMesssageAndFetchTransportReport(message, parameter);

			var result = client.fetchMessage(messageMetaData, parameter);

			assertThat(result).contains(transportReport);
		}
	}

	@DisplayName("process message and fetch transport report")
	@Nested
	class TestProcessMessageAndFetchTransportReport {

		@Mock
		private XtaMessageMetaData messageMetaData;

		@Mock
		private FetchMessageParameter parameter;

		@Mock
		private Consumer<XtaMessage> processMessageConsumer;

		@Mock
		private XtaMessage message;

		@Mock
		private XtaTransportReport transportReport;

		@BeforeEach
		void mock() {
			when(message.metaData()).thenReturn(messageMetaData);
			when(parameter.processMessage()).thenReturn(processMessageConsumer);
			doNothing().when(client).processMessageAndCloseIfNoException(any(), any());
			when(service.getTransportReport(messageMetaData)).thenReturn(Optional.of(transportReport));
		}

		@DisplayName("should call process message and close if no exception")
		@Test
		void shouldCallProcessMessageAndCloseIfNoException() {
			client.processMesssageAndFetchTransportReport(message, parameter);

			verify(client).processMessageAndCloseIfNoException(message, processMessageConsumer);
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = client.processMesssageAndFetchTransportReport(message, parameter);

			assertThat(result).contains(transportReport);
		}
	}

	@DisplayName("process message and close if no exception")
	@Nested
	class TestProcessMessageAndCloseIfNoException {

		@Mock
		private Consumer<XtaMessage> processMessageConsumer;

		@Mock
		private XtaMessage message;

		@BeforeEach
		void mock() {
			when(message.metaData()).thenReturn(listing.messages().getFirst());
		}

		@DisplayName("should consume message")
		@Test
		void shouldConsumeMessage() {
			processMessageAndCloseIfNoException();

			verify(processMessageConsumer).accept(message);
		}

		@DisplayName("should close message")
		@Test
		void shouldCloseMessage() {
			processMessageAndCloseIfNoException();

			verify(service).closeMessage(message);
		}

		@DisplayName("with runtime exception")
		@Nested
		class TestWithRuntimeException {

			@Mock
			private RuntimeException exception;

			@BeforeEach
			void mock() {
				doThrow(exception).when(processMessageConsumer).accept(message);
			}

			@DisplayName("should log error")
			@Test
			void shouldLogError() {
				processMessageAndCloseIfNoException();

				verify(client).logErrorForMessageProcessingFailure(MESSAGE_ID, exception);
			}

			@DisplayName("should not close message")
			@Test
			void shouldNotCloseMessage() {
				processMessageAndCloseIfNoException();

				verify(service, times(0)).closeMessage(any());
			}
		}

		private void processMessageAndCloseIfNoException() {
			client.processMessageAndCloseIfNoException(message, processMessageConsumer);
		}
	}

	@DisplayName("send message")
	@Nested
	class TestSendMessage {

		@Mock
		private XtaTransportReport transportReport;

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			doReturn(transportReport).when(client).sendMessageRaw(message);

			var result = client.sendMessage(message);

			assertThat(result).isEqualTo(transportReport);
		}

		@DisplayName("should throw checked exception on runtime exception")
		@Test
		@SneakyThrows
		void shouldThrowCheckedExceptionOnRuntimeException() {
			var clientException = XtaClientExceptionTestFactory.create();
			var runtimeException = ClientRuntimeExceptionTestFactory.create();
			doThrow(runtimeException).when(client).sendMessageRaw(message);
			when(exceptionHandler.deriveXtaClientException(runtimeException)).thenReturn(clientException);

			assertThatThrownBy(() -> client.sendMessage(message))
					.isEqualTo(clientException);
		}
	}

	@DisplayName("send message raw")
	@Nested
	class TestSendMessageRaw {

		@Mock
		private XtaTransportReport transportReport;

		@BeforeEach
		void mock() {
			doNothing().when(client).throwExceptionIfAccountInactive(any());
			doNothing().when(client).throwExceptionIfServiceNotAvailable(any());
			when(service.sendMessage(message)).thenReturn(transportReport);
		}

		@DisplayName("should call checkAccountActive")
		@Test
		@SneakyThrows
		void shouldCallCheckAccountActive() {
			sendMessage();

			verify(client).throwExceptionIfAccountInactive(AUTHOR_IDENTIFIER);
		}

		@DisplayName("should call lookup service")
		@Test
		@SneakyThrows
		void shouldCallLookupService() {
			sendMessage();

			verify(client).throwExceptionIfServiceNotAvailable(message.metaData());
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = sendMessage();

			assertThat(result).isEqualTo(transportReport);
		}

		@SneakyThrows
		private XtaTransportReport sendMessage() {
			return client.sendMessage(message);
		}
	}

	@DisplayName("throw exception if service not available")
	@Nested
	class TestThrowExceptionIfServiceNotAvailable {

		@DisplayName("should call lookupService")
		@Test
		void shouldCallLookupService() {
			doReturn(true).when(service).lookupService(any());

			client.throwExceptionIfServiceNotAvailable(message.metaData());

			verify(service).lookupService(message.metaData());
		}

		@DisplayName("should throw exception if service not available")
		@Test
		void shouldThrowExceptionIfServiceNotAvailable() {
			var metaData = message.metaData();
			doReturn(false).when(service).lookupService(message.metaData());

			assertThatThrownBy(() -> client.throwExceptionIfServiceNotAvailable(metaData))
					.isInstanceOf(XtaClientRuntimeException.class);
		}
	}

	@DisplayName("throw exception if account inactive")
	@Nested
	class TestThrowExceptionIfAccountInactive {

		@DisplayName("should call checkAccountActive")
		@Test
		void shouldCallCheckAccountActive() {
			doReturn(true).when(service).checkAccountActive(any());

			client.throwExceptionIfAccountInactive(SELF_IDENTIFIER);

			verify(service).checkAccountActive(SELF_IDENTIFIER);
		}

		@DisplayName("should throw exception if account inactive")
		@Test
		void shouldThrowExceptionIfAccountInactive() {
			doReturn(false).when(service).checkAccountActive(AUTHOR_IDENTIFIER);

			assertThatThrownBy(() -> client.throwExceptionIfAccountInactive(AUTHOR_IDENTIFIER))
					.isInstanceOf(XtaClientRuntimeException.class);
		}
	}

}