package de.ozgcloud.xta.client.config;

import static org.assertj.core.api.Assertions.*;

import java.util.List;
import java.util.function.UnaryOperator;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;

import de.ozgcloud.xta.client.exception.XtaClientInitializationException;
import de.ozgcloud.xta.client.factory.XtaClientConfigTestFactory;
import de.ozgcloud.xta.client.model.XtaIdentifier;
import lombok.SneakyThrows;

class XtaConfigValidatorTest {

	@InjectMocks
	private XtaConfigValidator validator;

	@DisplayName("validate")
	@Nested
	class TestValidate {

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			var config = XtaClientConfigTestFactory.create();

			validator.validate(config);
		}

		@DisplayName("should return with client keystore")
		@Test
		@SneakyThrows
		void shouldReturnWithClientKeystore() {
			var config = XtaClientConfigTestFactory.createBuilder()
					.clientCertKeystore(XtaClientConfigTestFactory.createKeystoreBuilder().build())
					.build();

			validator.validate(config);
		}

		@DisplayName("should throw with blank client identifier")
		@Test
		void shouldThrowWithBlankClientIdentifier() {
			var config = XtaClientConfigTestFactory.createBuilder()
					.clientIdentifiers(List.of(XtaIdentifier.builder().value("").build()))
					.build();

			assertThatThrownBy(() -> validator.validate(config))
					.isInstanceOf(XtaClientInitializationException.class)
					.hasMessageContaining("clientIdentifiers");
		}

		@DisplayName("should throw without management service url")
		@Test
		void shouldThrowWithoutManagementServiceUrl() {
			var config = XtaClientConfigTestFactory.createBuilder()
					.managementServiceUrl("")
					.build();

			assertThatThrownBy(() -> validator.validate(config))
					.isInstanceOf(XtaClientInitializationException.class)
					.hasMessageContaining("managementServiceUrl");
		}

		@DisplayName("should throw without send service url")
		@Test
		void shouldThrowWithoutSendServiceUrl() {
			var config = XtaClientConfigTestFactory.createBuilder()
					.sendServiceUrl("")
					.build();

			assertThatThrownBy(() -> validator.validate(config))
					.isInstanceOf(XtaClientInitializationException.class)
					.hasMessageContaining("sendServiceUrl");
		}

		@DisplayName("should throw without msg box service url")
		@Test
		void shouldThrowWithoutMsgBoxServiceUrl() {
			var config = XtaClientConfigTestFactory.createBuilder()
					.msgBoxServiceUrl("")
					.build();

			assertThatThrownBy(() -> validator.validate(config))
					.isInstanceOf(XtaClientInitializationException.class)
					.hasMessageContaining("msgBoxServiceUrl");
		}

		@DisplayName("should throw with non-positive max list items")
		@Test
		void shouldThrowWithNonPositiveMaxListItems() {
			var config = XtaClientConfigTestFactory.createBuilder()
					.maxListItems(0)
					.build();

			assertThatThrownBy(() -> validator.validate(config))
					.isInstanceOf(XtaClientInitializationException.class)
					.hasMessageContaining("maxListItems");
		}

		@DisplayName("with invalid client cert keystore")
		@Nested
		class TestWithInvalidClientCertKeystore {
			@DisplayName("should throw without password")
			@Test
			void shouldThrowWithoutKeystorePassword() {
				var config = createKeystoreWithClientCertKeyStore(o -> o.password(null));

				assertThatThrownBy(() -> validator.validate(config))
						.isInstanceOf(XtaClientInitializationException.class)
						.hasMessageContaining("password");
			}

			@DisplayName("should throw without type")
			@Test
			void shouldThrowWithoutKeystoreType() {
				var config = createKeystoreWithClientCertKeyStore(o -> o.type(""));

				assertThatThrownBy(() -> validator.validate(config))
						.isInstanceOf(XtaClientInitializationException.class)
						.hasMessageContaining("type");
			}

			@DisplayName("should throw without content")
			@Test
			void shouldThrowWithoutKeystoreContent() {
				var config = createKeystoreWithClientCertKeyStore(o -> o.content(null));

				assertThatThrownBy(() -> validator.validate(config))
						.isInstanceOf(XtaClientInitializationException.class)
						.hasMessageContaining("content");
			}

			private XtaClientConfig createKeystoreWithClientCertKeyStore(UnaryOperator<XtaClientConfig.KeyStore.KeyStoreBuilder> operator) {
				return XtaClientConfigTestFactory.createBuilder()
						.clientCertKeystore(operator.apply(XtaClientConfigTestFactory.createKeystoreBuilder()).build())
						.build();
			}
		}

	}
}