package de.ozgcloud.xta.client.core;

import static de.ozgcloud.xta.client.core.XtaPortTripleFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jakarta.xml.ws.BindingProvider;

import org.apache.cxf.configuration.jsse.TLSClientParameters;
import org.apache.cxf.endpoint.Client;
import org.apache.cxf.ext.logging.LoggingInInterceptor;
import org.apache.cxf.ext.logging.LoggingOutInterceptor;
import org.apache.cxf.interceptor.Interceptor;
import org.apache.cxf.transport.Conduit;
import org.apache.cxf.transport.http.HTTPConduit;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.xta.client.factory.XtaClientConfigTestFactory;
import de.ozgcloud.xta.client.config.XtaClientConfig;
import genv3.de.xoev.transport.xta.x211.ManagementPortType;
import genv3.de.xoev.transport.xta.x211.MsgBoxPortType;
import genv3.de.xoev.transport.xta.x211.SendPortType;
import genv3.de.xoev.transport.xta.x211.XTAService;
import lombok.SneakyThrows;

class XtaPortTripleFactoryTest {
	private final String endpointUrl = "http://endpoint.test/test.svc";

	@Mock
	private XTAService xtaService;

	@Mock
	private XtaClientConfig config;

	@Mock
	private XtaTLSClientParametersFactory sslContextFactory;

	@Spy
	@InjectMocks
	private XtaPortTripleFactory factory;

	@DisplayName("create")
	@Nested
	class TestCreate {

		private final XtaClientConfig testConfig = XtaClientConfigTestFactory.create();

		@Mock(extraInterfaces = BindingProvider.class)
		private ManagementPortType managementPort;
		@Mock(extraInterfaces = BindingProvider.class)
		private SendPortType sendXtaPort;
		@Mock(extraInterfaces = BindingProvider.class)
		private MsgBoxPortType msgBoxPort;


		@BeforeEach
		@SneakyThrows
		void mock() {
			when(config.getManagementServiceUrl()).thenReturn(testConfig.getManagementServiceUrl());
			when(config.getSendServiceUrl()).thenReturn(testConfig.getSendServiceUrl());
			when(config.getMsgBoxServiceUrl()).thenReturn(testConfig.getMsgBoxServiceUrl());

			when(xtaService.getManagementPort()).thenReturn(managementPort);
			when(xtaService.getSendXtaPort()).thenReturn(sendXtaPort);
			when(xtaService.getMsgBoxPort()).thenReturn(msgBoxPort);

			doNothing().when(factory).configureBinding(any(), any());
		}

		@DisplayName("should configure management port")
		@Test
		@SneakyThrows
		void shouldConfigureManagementPort() {
			factory.create();

			verify(factory).configureBinding(testConfig.getManagementServiceUrl(), (BindingProvider) managementPort);
		}

		@DisplayName("should return management port")
		@Test
		@SneakyThrows
		void shouldReturnManagementPort() {
			var result = factory.create();

			assertThat(result.managementPort()).isEqualTo(managementPort);
		}

		@DisplayName("should configure send port")
		@Test
		@SneakyThrows
		void shouldConfigureSendPort() {
			factory.create();

			verify(factory).configureBinding(testConfig.getSendServiceUrl(), (BindingProvider) sendXtaPort);
		}

		@DisplayName("should return send port")
		@Test
		@SneakyThrows
		void shouldReturnSendPort() {
			var result = factory.create();

			assertThat(result.sendPort()).isEqualTo(sendXtaPort);
		}

		@DisplayName("should configure msg box port")
		@Test
		@SneakyThrows
		void shouldConfigureMsgBoxPort() {
			factory.create();

			verify(factory).configureBinding(testConfig.getMsgBoxServiceUrl(), (BindingProvider) msgBoxPort);
		}

		@DisplayName("should return msg box port")
		@Test
		@SneakyThrows
		void shouldReturnMsgBoxPort() {
			var result = factory.create();

			assertThat(result.msgBoxPort()).isEqualTo(msgBoxPort);
		}
	}

	@DisplayName("configure binding")
	@Nested
	class TestConfigureBinding {

		@Mock
		private BindingProvider port;

		@Mock
		private Map<String, Object> requestContext;

		@Mock
		private Client client;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(port.getRequestContext()).thenReturn(requestContext);
			doNothing().when(factory).configureClient(any());
			doReturn(client).when(factory).getClientFromPort(port);
		}

		@DisplayName("should configure request context")
		@Test
		@SneakyThrows
		void shouldConfigureRequestContext() {
			factory.configureBinding(endpointUrl, port);

			verify(factory).configureRequestContext(endpointUrl, requestContext);
		}

		@DisplayName("should configure client")
		@Test
		@SneakyThrows
		void shouldConfigureClient() {
			factory.configureBinding(endpointUrl, port);

			verify(factory).configureClient(client);
		}
	}

	@DisplayName("configure request context")
	@Nested
	class TestConfigureRequestContext {

		private HashMap<String, Object> requestContext;

		@BeforeEach
		void mock() {
			requestContext = new HashMap<>();
			requestContext.put("old", "value");
			when(config.isSchemaValidation()).thenReturn(false);
		}

		@DisplayName("should keep old")
		@Test
		@SneakyThrows
		void shouldKeepOld() {
			factory.configureRequestContext(endpointUrl, requestContext);

			assertThat(requestContext).containsEntry("old", "value");
		}

		@DisplayName("should add endpoint url")
		@Test
		@SneakyThrows
		void shouldAddEndpointUrl() {
			factory.configureRequestContext(endpointUrl, requestContext);

			assertThat(requestContext).containsEntry(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, endpointUrl);
		}

		@DisplayName("should add schema validation")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		@SneakyThrows
		void shouldAddSchemaValidation(boolean validation) {
			reset(config);
			when(config.isSchemaValidation()).thenReturn(validation);

			factory.configureRequestContext(endpointUrl, requestContext);

			assertThat(requestContext).containsEntry(SCHEMA_VALIDATION_ENABLED_KEY, validation);
		}
	}

	@DisplayName("configure client")
	@Nested
	class TestConfigureClient {

		@Mock
		private Client client;

		@Mock
		private List<Interceptor<?>> outInterceptors;

		@Mock
		private List<Interceptor<?>> inInterceptors;

		@Mock(extraInterfaces = Conduit.class)
		private HTTPConduit clientConduit;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(client.getConduit()).thenReturn(clientConduit);
			doNothing().when(factory).configureHttpConduit(any());
		}

		@DisplayName("should configure out interceptors")
		@Test
		@SneakyThrows
		void shouldConfigureOutInterceptors() {
			when(config.isLogSoapRequests()).thenReturn(true);
			when(config.isLogSoapResponses()).thenReturn(false);
			when(client.getOutInterceptors()).thenReturn(outInterceptors);
			doNothing().when(factory).configureOutInterceptors(any());

			factory.configureClient(client);

			verify(factory).configureOutInterceptors(client.getOutInterceptors());
		}

		@DisplayName("should configure in interceptors")
		@Test
		@SneakyThrows
		void shouldConfigureInInterceptors() {
			when(config.isLogSoapRequests()).thenReturn(false);
			when(config.isLogSoapResponses()).thenReturn(true);
			when(client.getInInterceptors()).thenReturn(inInterceptors);
			doNothing().when(factory).configureInInterceptors(any());

			factory.configureClient(client);

			verify(factory).configureInInterceptors(client.getInInterceptors());
		}

		@DisplayName("should configure http conduit")
		@Test
		@SneakyThrows
		void shouldConfigureHttpConduit() {
			when(config.isLogSoapRequests()).thenReturn(false);
			when(config.isLogSoapResponses()).thenReturn(false);

			factory.configureClient(client);

			verify(factory).configureHttpConduit((HTTPConduit) client.getConduit());
		}
	}

	@DisplayName("configure in interceptors")
	@Nested
	class TestConfigureInInterceptors {

		private List<Interceptor<?>> inInterceptors;

		@BeforeEach
		void setup() {
			inInterceptors = new ArrayList<>();
		}

		@DisplayName("should add one interceptor")
		@Test
		void shouldAddOneInterceptor() {
			factory.configureInInterceptors(inInterceptors);

			assertThat(inInterceptors).hasSize(1);
		}

		@DisplayName("should add in interceptor")
		@Test
		@SneakyThrows
		void shouldAddInInterceptor() {
			factory.configureInInterceptors(inInterceptors);

			assertThat(inInterceptors.getFirst()).isInstanceOf(LoggingInInterceptor.class);
		}
	}

	@DisplayName("configure out interceptors")
	@Nested
	class TestConfigureOutInterceptors {

		private List<Interceptor<?>> outInterceptors;

		@BeforeEach
		void setup() {
			outInterceptors = new ArrayList<>();
		}

		@DisplayName("should add one interceptor")
		@Test
		void shouldAddOneInterceptor() {
			factory.configureOutInterceptors(outInterceptors);

			assertThat(outInterceptors).hasSize(1);
		}

		@DisplayName("should add out interceptor")
		@Test
		@SneakyThrows
		void shouldAddOutInterceptor() {
			factory.configureOutInterceptors(outInterceptors);

			assertThat(outInterceptors.getFirst()).isInstanceOf(LoggingOutInterceptor.class);
		}
	}

	@DisplayName("configure http conduit")
	@Nested
	class TestConfigureHttpConduit {

		@Mock
		private HTTPConduit conduit;

		@Mock
		private TLSClientParameters tlsClientParameters;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(sslContextFactory.create()).thenReturn(tlsClientParameters);
		}

		@DisplayName("should set tls client parameters")
		@Test
		@SneakyThrows
		void shouldSetTlsClientParameters() {
			factory.configureHttpConduit(conduit);

			verify(conduit).setTlsClientParameters(tlsClientParameters);
		}
	}

}