package de.ozgcloud.xta.client.mapper;

import static de.ozgcloud.xta.client.factory.MessageMetaDataTestFactory.*;
import static de.ozgcloud.xta.client.factory.XtaClientConfigTestFactory.*;
import static de.ozgcloud.xta.client.factory.XtaFileTestFactory.*;
import static de.ozgcloud.xta.client.mapper.RequestMapper.*;
import static org.assertj.core.api.Assertions.*;

import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;

import de.ozgcloud.xta.client.factory.XtaFileTestFactory;
import de.ozgcloud.xta.client.factory.XtaMessageMetaDataTestFactory;
import de.ozgcloud.xta.client.factory.XtaMessageTestFactory;
import de.ozgcloud.xta.client.model.XtaFile;
import de.ozgcloud.xta.client.model.XtaMessage;
import genv3.de.xoev.transport.xta.x211.ContentType;
import genv3.de.xoev.transport.xta.x211.LookupServiceType;
import genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData;
import genv3.eu.osci.ws.x2014.x10.transport.PartyIdentifierType;
import lombok.SneakyThrows;

class RequestMapperTest {

	private final RequestMapper mapper = Mappers.getMapper(RequestMapper.class);

	@DisplayName("map party type from identifier")
	@Nested
	class TestMapPartyTypeFromIdentifier {

		@DisplayName("should use value")
		@Test
		void shouldUseValue() {
			var result = mapPartyIdentifier();

			assertThat(result.getValue()).isEqualTo(SELF_IDENTIFIER.value());
		}

		@DisplayName("should use category")
		@Test
		void shouldUseCategory() {
			var result = mapPartyIdentifier();

			assertThat(result.getCategory()).isEqualTo(SELF_IDENTIFIER.category());
		}

		@DisplayName("should use name")
		@Test
		void shouldUseName() {
			var result = mapPartyIdentifier();

			assertThat(result.getName()).isEqualTo(SELF_IDENTIFIER.name());
		}

		@DisplayName("should use type")
		@Test
		void shouldUseType() {
			var result = mapPartyIdentifier();

			assertThat(result.getType()).isEqualTo(IDENTIFIER_TYPE);
		}

		private PartyIdentifierType mapPartyIdentifier() {
			var result = mapper.mapPartyTypeFromIdentifier(SELF_IDENTIFIER);

			return result.getIdentifier();
		}
	}

	@DisplayName("map msg box status list request from max list items")
	@Nested
	class TestMapMsgBoxStatusListRequestFromMaxListItems {

		@DisplayName("should use max list items")
		@Test
		void shouldUseMaxListItems() {
			var result = mapper.mapMsgBoxStatusListRequestTypeFromMaxListItems(MAX_LIST_ITEMS);

			assertThat(result.getMaxListItems().intValue()).isEqualTo(MAX_LIST_ITEMS);
		}
	}

	@DisplayName("map message meta data from xta message meta data")
	@Nested
	class TestMapMessageMetaDataFromXtaMessageMetaData {

		@DisplayName("should map service")
		@Test
		void shouldMapService() {
			var result = map();

			assertThat(result.getQualifier().getService()).isEqualTo(MESSAGE_SERVICE);
		}

		@DisplayName("should map business scenario defined listURI")
		@Test
		void shouldMapBusinessScenarioDefinedListUri() {
			var result = map();

			assertThat(result.getQualifier().getBusinessScenario().getDefined().getListURI())
					.isEqualTo(BUSINESS_SCENARIO_LIST_URI);
		}

		@DisplayName("should map business scenario defined listVersionID")
		@Test
		void shouldMapBusinessScenarioDefinedListVersionID() {
			var result = map();

			assertThat(result.getQualifier().getBusinessScenario().getDefined().getListVersionID())
					.isEqualTo(BUSINESS_SCENARIO_LIST_VERSION_ID);
		}

		@DisplayName("should map business scenario code")
		@Test
		void shouldMapBusinessScenarioCode() {
			var result = map();

			assertThat(result.getQualifier().getBusinessScenario().getDefined().getCode())
					.isEqualTo(BUSINESS_SCENARIO_CODE);
		}

		@DisplayName("should map message type code")
		@Test
		void shouldMapMessageTypeCode() {
			var result = map();

			assertThat(result.getQualifier().getMessageType().getCode()).isEqualTo(MESSAGE_TYPE_CODE);
		}

		@DisplayName("should map message type payload schema")
		@Test
		void shouldMapMessageTypePayloadSchema() {
			var result = map();

			assertThat(result.getQualifier().getMessageType().getPayloadSchema())
					.isEqualTo(MESSAGE_TYPE_PAYLOAD_SCHEMA);
		}

		@DisplayName("should map message type listURI")
		@Test
		void shouldMapMessageTypeListUri() {
			var result = map();

			assertThat(result.getQualifier().getMessageType().getListURI()).isEqualTo(MESSAGE_TYPE_LIST_URI);
		}

		@DisplayName("should map message type listVersionID")
		@Test
		void shouldMapMessageTypeListVersionId() {
			var result = map();

			assertThat(result.getQualifier().getMessageType().getListVersionID()).isEqualTo(MESSAGE_TYPE_LIST_VERSION_ID);
		}

		@DisplayName("should map message id")
		@Test
		void shouldMapMessageId() {
			var result = map();

			assertThat(result.getMsgIdentification().getMessageID().getValue()).isEqualTo(MESSAGE_ID);
		}

		@DisplayName("should map author identifier")
		@Test
		void shouldMapAuthorIdentifier() {
			var result = map();

			assertThat(result.getOriginators().getAuthor().getIdentifier()).usingRecursiveComparison().isEqualTo(
					mapper.mapPartyIdentifierTypeFromIdentifier(AUTHOR_IDENTIFIER));
		}

		@DisplayName("should map reader identifier")
		@Test
		void shouldMapReaderIdentifier() {
			var result = map();

			assertThat(result.getDestinations().getReader().getIdentifier()).usingRecursiveComparison().isEqualTo(
					mapper.mapPartyIdentifierTypeFromIdentifier(READER_IDENTIFIER));
		}

		@DisplayName("with undefined business scenario")
		@Nested
		class TestWithUndefinedBusinessScenario {
			@DisplayName("should map code")
			@Test
			void shouldMapCode() {
				var result = mapWithUndefinedBusinessScenario();

				assertThat(result.getQualifier().getBusinessScenario().getUndefined()).isEqualTo(BUSINESS_SCENARIO_CODE);
			}

			@DisplayName("should map defined to null")
			@Test
			void shouldMapDefinedToNull() {
				var result = mapWithUndefinedBusinessScenario();

				assertThat(result.getQualifier().getBusinessScenario().getDefined()).isNull();
			}

			private MessageMetaData mapWithUndefinedBusinessScenario() {
				var metaData = XtaMessageMetaDataTestFactory.createBuilder()
						.businessScenarioListUri(null)
						.build();

				return mapper.mapMessageMetaDataFromXtaMessageMetaData(metaData);
			}
		}

		private MessageMetaData map() {
			return mapper.mapMessageMetaDataFromXtaMessageMetaData(XtaMessageMetaDataTestFactory.create());
		}
	}

	@DisplayName("map generic content container")
	@Nested
	class TestMapGenericContentContainer {
		private XtaMessage xtaMessage;

		@BeforeEach
		void setup() {
			xtaMessage = XtaMessageTestFactory.create();
		}

		@DisplayName("should map message")
		@Test
		@SneakyThrows
		void shouldMapMessage() {
			var result = mapper.mapGenericContentContainer(xtaMessage);

			var messageFileId = result.getContentContainer().getMessage().getId();
			assertThat(messageFileId).isEqualTo(XTA_FILE_ID);
		}

		@DisplayName("should map attachments")
		@Test
		@SneakyThrows
		void shouldMapAttachments() {
			var result = mapper.mapGenericContentContainer(xtaMessage);

			var attachmentFileIds = result.getContentContainer().getAttachment().stream()
					.map(ContentType::getId)
					.toList();
			assertThat(attachmentFileIds).containsExactly(XTA_FILE_ID_2, XTA_FILE_ID_3);
		}

		@DisplayName("should set encrypted data to null")
		@Test
		void shouldSetEncryptedDataToNull() {
			var result = mapper.mapGenericContentContainer(xtaMessage);

			assertThat(result.getEncryptedData()).isNull();
		}

	}

	@DisplayName("map content type")
	@Nested
	class TestMapContentType {
		private final XtaFile xtaFile = XtaFileTestFactory.create();

		@DisplayName("should map value")
		@Test
		@SneakyThrows
		void shouldMapValue() {
			var result = doMapping();

			assertThat(IOUtils.toByteArray(result.getValue().getInputStream())).isEqualTo(XTA_FILE_CONTENT);
		}

		@DisplayName("should map content description")
		@Test
		void shouldMapContentDescription() {
			var result = doMapping();

			assertThat(result.getContentDescription()).isEqualTo(XTA_FILE_CONTENT_DESCRIPTION);
		}

		@DisplayName("should map content type")
		@Test
		void shouldMapContentType() {
			var result = doMapping();

			assertThat(result.getContentType()).isEqualTo(XTA_FILE_CONTENT_TYPE);
		}

		@DisplayName("should map encoding")
		@Test
		void shouldMapEncoding() {
			var result = doMapping();

			assertThat(result.getEncoding()).isEqualTo(XTA_FILE_ENCODING);
		}

		@DisplayName("should map filename")
		@Test
		void shouldMapFilename() {
			var result = doMapping();

			assertThat(result.getFilename()).isEqualTo(XTA_FILE_NAME);
		}

		@DisplayName("should map id")
		@Test
		void shouldMapId() {
			var result = doMapping();

			assertThat(result.getId()).isEqualTo(XTA_FILE_ID);
		}

		@DisplayName("should map lang")
		@Test
		void shouldMapLang() {
			var result = doMapping();

			assertThat(result.getLang()).isEqualTo(XTA_FILE_LANGUAGE);
		}

		@DisplayName("should map size")
		@Test
		void shouldMapSize() {
			var result = doMapping();

			assertThat(result.getSize()).isEqualTo(XTA_FILE_SIZE);
		}

		private ContentType doMapping() {
			return mapper.mapContentType(xtaFile);
		}
	}

	@DisplayName("map msg box close request type")
	@Nested
	class TestMapMsgBoxCloseRequestType {
		@DisplayName("should map message id")
		@Test
		void shouldMapMessageId() {
			var result = mapper.mapMsgBoxCloseRequestType(MESSAGE_ID);

			assertThat(result.getLastMsgReceived().getFirst().getValue()).isEqualTo(MESSAGE_ID);
		}

		@DisplayName("should map request id")
		@Test
		void shouldMapRequestId() {
			var result = mapper.mapMsgBoxCloseRequestType(MESSAGE_ID);

			assertThat(result.getMsgBoxRequestID()).isEqualTo(CLOSE_REQUEST_ID);
		}
	}

	@DisplayName("map lookup service request")
	@Nested
	class TestMapLookupServiceRequest {
		@DisplayName("should map service")
		@Test
		void shouldMapServiceAndReader() {
			var result = getFirstLookupServiceType();

			assertThat(result.getServiceType()).isEqualTo(MESSAGE_SERVICE);
		}

		@DisplayName("should map reader")
		@Test
		void shouldMapReader() {
			var result = getFirstLookupServiceType();

			assertThat(result.getReader().getIdentifier()).usingRecursiveComparison().isEqualTo(
					mapper.mapPartyIdentifierTypeFromIdentifier(READER_IDENTIFIER));
		}

		LookupServiceType getFirstLookupServiceType() {
			var result = mapper.mapLookupServiceRequest(MESSAGE_SERVICE, READER_IDENTIFIER);

			return result.getLookupServiceRequestList().getFirst().getLookupService();
		}

	}

	@DisplayName("map msg box fetch request")
	@Nested
	class TestMapMsgBoxFetchRequest {
		@DisplayName("should map message id")
		@Test
		void shouldMapMessageId() {
			var result = mapper.mapMsgBoxFetchRequest(MESSAGE_ID);

			assertThat(result.getMsgSelector().getMessageID().getFirst().getValue()).isEqualTo(MESSAGE_ID);
		}
	}

}