package de.ozgcloud.xta.client.xdomea.reader;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.w3c.dom.Document;

import de.ozgcloud.xta.client.exception.XtaClientException;
import lombok.SneakyThrows;

class XdomeaValueReaderTest {

	@InjectMocks
	@Spy
	private XdomeaValueReader xdomeaValueReader;

	@Mock
	private Map<String, XmlValueReader> xmlValueReaders;

	private static final String XPATH_STRING = "xpath/expression";

	@DisplayName("read values")
	@Nested
	class TestReadValues {

		private static final String PROCESS_ID = "process id";
		private static final String MESSAGE_TYPE_CODE = "message type code";
		private static final String AUTHOR_ID_SUFFIX = "author id suffix";
		private static final String AUTHOR_ID_PREFIX = "author id prefix";
		private static final String READER_ID_SUFFIX = "reader id suffix";
		private static final String READER_ID_PREFIX = "reader id prefix";
		private static final List<String> PRIMARY_DOCUMENT_NAMES = List.of("primary document name");

		@Mock
		Document xdomeaXmlDocument;

		@BeforeEach
		@SneakyThrows
		void mock() {
			doReturn(PROCESS_ID).when(xdomeaValueReader).readRequiredValue(xdomeaXmlDocument, XdomeaValueReader.PROCESS_ID_XPATH);
			doReturn(MESSAGE_TYPE_CODE).when(xdomeaValueReader).readRequiredValue(xdomeaXmlDocument, XdomeaValueReader.MESSAGE_TYPE_ID_SUFFIX_XPATH);
			doReturn(AUTHOR_ID_SUFFIX).when(xdomeaValueReader).readRequiredValue(xdomeaXmlDocument, XdomeaValueReader.AUTHOR_ID_SUFFIX_XPATH);
			doReturn(AUTHOR_ID_PREFIX).when(xdomeaValueReader).readRequiredValue(xdomeaXmlDocument, XdomeaValueReader.AUTHOR_ID_PREFIX_XPATH);
			doReturn(READER_ID_SUFFIX).when(xdomeaValueReader).readRequiredValue(xdomeaXmlDocument, XdomeaValueReader.READER_ID_SUFFIX_XPATH);
			doReturn(READER_ID_PREFIX).when(xdomeaValueReader).readRequiredValue(xdomeaXmlDocument, XdomeaValueReader.READER_ID_PREFIX_XPATH);
			doReturn(PRIMARY_DOCUMENT_NAMES).when(xdomeaValueReader).readPrimaryDocumentNames(xdomeaXmlDocument);
		}

		@DisplayName("should return process id")
		@Test
		void shouldReturnProcessId() {
			var result = readValues();

			assertThat(result.processId()).isEqualTo(PROCESS_ID);
		}

		@DisplayName("should return message type code")
		@Test
		void shouldReturnMessageTypeCode() {
			var result = readValues();

			assertThat(result.messageTypeCode()).isEqualTo(MESSAGE_TYPE_CODE);
		}

		@DisplayName("should return author id suffix")
		@Test
		void shouldReturnAuthorIdSuffix() {
			var result = readValues();

			assertThat(result.authorIdSuffix()).isEqualTo(AUTHOR_ID_SUFFIX);
		}

		@DisplayName("should return author id prefix")
		@Test
		void shouldReturnAuthorIdPrefix() {
			var result = readValues();

			assertThat(result.authorIdPrefix()).isEqualTo(AUTHOR_ID_PREFIX);
		}

		@DisplayName("should return reader id suffix")
		@Test
		void shouldReturnReaderIdSuffix() {
			var result = readValues();

			assertThat(result.readerIdSuffix()).isEqualTo(READER_ID_SUFFIX);
		}

		@DisplayName("should return reader id prefix")
		@Test
		void shouldReturnReaderIdPrefix() {
			var result = readValues();

			assertThat(result.readerIdPrefix()).isEqualTo(READER_ID_PREFIX);
		}

		@DisplayName("should return primary document names")
		@Test
		void shouldReturnPrimaryDocumentNames() {
			var result = readValues();

			assertThat(result.primaryDocumentNames()).isEqualTo(PRIMARY_DOCUMENT_NAMES);
		}

		@SneakyThrows
		private XdomeaXmlValues readValues() {
			return xdomeaValueReader.readValues(xdomeaXmlDocument);
		}
	}

	@DisplayName("read primary document names")
	@Nested
	class TestReadPrimaryDocumentNames {

		@Mock
		XmlValueReader xmlValueReader;

		@Mock
		Document xdomeaXmlDocument;

		final String documentName1 = "dou.pdf";
		final String documentName2 = "doc.png";

		@BeforeEach
		void mock() {
			doReturn(xmlValueReader).when(xdomeaValueReader).getXmlValueReader(XdomeaValueReader.PRIMARY_DOCUMENT_NAME_XPATH);
		}

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			when(xmlValueReader.readNonEmptyTexts(xdomeaXmlDocument)).thenReturn(Stream.of(documentName1, documentName2));

			var result = xdomeaValueReader.readPrimaryDocumentNames(xdomeaXmlDocument);

			assertThat(result).containsExactly(documentName1, documentName2);
		}
	}

	@DisplayName("read required value")
	@Nested
	class TestReadRequiredValue {

		@Mock
		XmlValueReader xmlValueReader;

		@Mock
		Document xdomeaXmlDocument;

		final String foundValue = "found value";

		@BeforeEach
		void mock() {
			doReturn(xmlValueReader).when(xdomeaValueReader).getXmlValueReader(XPATH_STRING);
		}

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			when(xmlValueReader.readNonEmptyTexts(xdomeaXmlDocument)).thenReturn(Stream.of(foundValue));

			var result = xdomeaValueReader.readRequiredValue(xdomeaXmlDocument, XPATH_STRING);

			assertThat(result).isEqualTo(foundValue);
		}

		@DisplayName("should throw client exception if missing")
		@Test
		void shouldThrowClientExceptionIfMissing() {
			when(xmlValueReader.readNonEmptyTexts(xdomeaXmlDocument)).thenReturn(Stream.empty());

			assertThatThrownBy(() -> xdomeaValueReader.readRequiredValue(xdomeaXmlDocument, XPATH_STRING))
					.isInstanceOf(XtaClientException.class);
		}
	}

	@DisplayName("get xml value reader")
	@Nested
	class TestGetXmlValueReader {

		@Mock
		XmlValueReader xmlValueReader;

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			when(xmlValueReaders.get(XPATH_STRING)).thenReturn(xmlValueReader);

			var result = xdomeaValueReader.getXmlValueReader(XPATH_STRING);

			assertThat(result).isEqualTo(xmlValueReader);
		}
	}

}