package de.ozgcloud.xta.client.xdomea.reader;

import static de.ozgcloud.xta.client.factory.XtaMessageTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.function.Function;

import jakarta.activation.DataHandler;
import jakarta.activation.FileDataSource;
import jakarta.mail.util.ByteArrayDataSource;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;

class ZipFileEntryReaderTest {
	private static final DataHandler ZIP_FILE = EXAMPLE_ABGABE_MESSAGE.messageFile().content();
	private static final String ENTRY_NAME = "84be3140-e069-4d4a-80c7-889b5eec539e_Abgabe.Abgabe.0401.xml";

	private static final DataHandler BAD_ZIP_FILE = createContentDataHandler("bad zip file".getBytes());

	private final ZipFileEntryReader reader = ZipFileEntryReaderFactory.builder().build().create();

	@DisplayName("map entry data")
	@Nested
	class TestMapEntryData {

		private final Function<InputStream, String> inputStreamConsumer = inputStream -> {
			try {
				var data = inputStream.readAllBytes();
				return new String(data);
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
		};

		@DisplayName("should map")
		@Test
		void shouldMap() {
			var xmlString = reader.mapEntryData(ZIP_FILE, ENTRY_NAME, inputStreamConsumer);

			assertThat(xmlString).contains("xdomea:Abgabe.Abgabe.0401");
		}

		@DisplayName("should throw exception if entry is missing")
		@Test
		void shouldThrowExceptionIfEntryIsMissing() {
			assertThatThrownBy(() -> reader.mapEntryData(ZIP_FILE, "?", inputStreamConsumer))
					.isInstanceOf(XtaClientRuntimeException.class);
		}

		@DisplayName("should throw exception on bad zip file")
		@Test
		void shouldThrowExceptionOnBadZipFile() {
			assertThatThrownBy(() -> reader.mapEntryData(BAD_ZIP_FILE, ENTRY_NAME, inputStreamConsumer))
					.isInstanceOf(XtaClientRuntimeException.class);
		}
	}

	@DisplayName("get entry names")
	@Nested
	class TestGetEntryNames {

		@DisplayName("should return entry names")
		@Test
		void shouldReturnEntryNames() {
			var entryNames = reader.getEntryNames(ZIP_FILE);

			assertThat(entryNames).containsExactlyInAnyOrder(
					"66a7635a-3872-3519-5ad8-00000000e7db_small_png.png",
					"66a76331-3872-3519-5ad8-00000000e7d8_XML-Daten.xml",
					"66a76372-3872-3519-5ad8-00000000e7df_Tempo.pdf",
					"84be3140-e069-4d4a-80c7-889b5eec539e_Abgabe.Abgabe.0401.xml"
			);
		}

		@DisplayName("should throw exception on bad zip file")
		@Test
		void shouldThrowExceptionOnBadZipFile() {
			assertThatThrownBy(() -> reader.getEntryNames(createMissingFileDataHandler()))
					.isInstanceOf(XtaClientRuntimeException.class);
		}

		private static DataHandler createMissingFileDataHandler() {
			return new DataHandler(new FileDataSource(new File("missing-file")));
		}
	}

	private static DataHandler createContentDataHandler(byte[] data) {
		return new DataHandler(new ByteArrayDataSource(data, "application/octet-stream"));
	}

}